"""
Test the environment variable check class
"""
from concurrent.futures import ThreadPoolExecutor

import pytest

from geomancy.checks import Check, Result, CheckException


class CheckDummy(Check):
    """A dummy Check subclass for testing aliases"""

    aliases = ("checkDummy",)


class HangCheck(Check):
    """A Check subclass that allows thread hanging"""

    locked = True

    def check(self, executor, level: int = 0):
        # The following hangs the thread until self.locked is set to False
        while self.locked:
            pass
        return Result(msg=self.name, status="passed")


class DefaultCheck(Check):
    """A Check subclass that reads the returned result status from an instance
    attribute"""

    default_status = "passed"

    def check(self, executor, level: int = 0):
        return Result(msg=self.name, status=self.default_status)


def test_check_init_env_subtitute():
    """Test the Check init with env_substitute specified"""
    for kwargs, expected in (
        ({"env_substitute": False}, False),
        ({"substitute": False}, False),
        ({}, True),  # default
    ):
        check = Check(name="EnvTest", **kwargs)
        assert check.env_substitute == expected


def test_check_init_condition():
    """Test the Check init with condition specified"""
    # 2. Test condition parameter aliases
    for kwargs, expected in (
        ({"condition": "all"}, all),
        ({"condition": "any"}, any),
        ({"subchecks": "all"}, all),
        ({"subchecks": "any"}, any),
        ({}, all),  # default
    ):
        check = Check(name="ConditionTest", **kwargs)
        assert check.condition == expected


def test_check_init_bad_child():
    """Test the Check init with a non-Check child"""
    with pytest.raises(CheckException):
        Check(name="BadChild", children=["a"])


def test_check_env_substitution():
    """Test the Check environment substitution of the value property"""
    with pytest.MonkeyPatch.context() as mp:
        # Set some present and missing environment variables
        mp.setenv("VARIABLE", "my value")
        mp.delenv("MISSING", raising=False)

        # 1. Present environment variable without braces
        assert Check(name="sub_env", value="$VARIABLE").raw_value == "$VARIABLE"
        assert Check(name="sub_env", value="$VARIABLE").value == "my value"

        # 2. Present environment variable with braces
        assert Check(name="sub_env", value="${VARIABLE}").raw_value == "${VARIABLE}"
        assert Check(name="sub_env", value="${VARIABLE}").value == "my value"

        # 3. Missing environment variable with braces substitute to empty string
        assert Check(name="sub_env", value="${MISSING}").raw_value == "${MISSING}"
        assert Check(name="sub_env", value="${MISSING}").value == ""

        # 4. Disable substitution
        #   4.1. Use the env_substitute=False flag without brace
        assert (
            Check(name="sub_env", value="$VARIABLE", env_substitute=False).value
            == "$VARIABLE"
        )

        #   4.2. Use the substitute=False flag without brace
        assert (
            Check(name="sub_env", value="$VARIABLE", substitute=False).value
            == "$VARIABLE"
        )

        #   4.3. Use the substitute=False flag with brace
        assert (
            Check(name="sub_env", value="${VARIABLE}", substitute=False).value
            == "${VARIABLE}"
        )

        #   4.4. Use string literal with and without brace
        assert Check(name="sub_env", value="'$VARIABLE'").value == "$VARIABLE"
        assert Check(name="sub_env", value="'${VARIABLE}'").value == "${VARIABLE}"

        #   4.5. Use the substitute=False flag without brace for missing variable
        assert (
            Check(name="sub_env", value="$MISSING", substitute=False).value
            == "$MISSING"
        )

        #   4.6. Use string literal for missing variable
        assert Check(name="sub_env", value="'$MISSING'").value == "$MISSING"


# noinspection GrazieInspection
def test_check_flatten_count():
    """Test the Check.flatten and Check.count properties"""
    # Create a Check tree
    sub11 = Check(name="sub11")
    sub12 = Check(name="sub12")
    sub1 = Check(name="sub1", children=[sub11, sub12])
    sub2 = Check(name="sub2")
    root = Check(name="root", children=[sub1, sub2])

    # Check the flattened tree
    flattened = root.flatten

    assert len(flattened) == 5
    assert root.count == 5
    assert flattened[0] == root
    assert flattened[1] == sub1
    assert flattened[2] == sub11
    assert flattened[3] == sub12
    assert flattened[4] == sub2


def test_check_types():
    """Test the Check.types() method"""

    types = Check.types()

    # All keys should be strings (class names)
    assert all(isinstance(k, str) for k in types.keys())

    # All values should be classes
    assert all(issubclass(v, Check) or v == Check for v in types.values())


# noinspection GrazieInspection
def test_check_load_simple():
    """Test the Check.load method from a simple dict"""
    d = {"CheckDummy": "PATH", "desc": "The $PATH environment variable"}

    check = Check.load(d=d, name="base check")

    assert check is not None
    assert isinstance(check, Check)
    assert check.__class__ == CheckDummy
    assert check.value == "PATH"
    assert check.desc == d["desc"]


def test_check_load_nested():
    """Test the Check.load method from a nested dict"""
    d = {
        "Environment": {
            "desc": "Environment checks",
            "Path": {"checkDummy": "VAR1", "desc": "Check VAR2"},
            "Term": {"checkDummy": "VAR2", "desc": "Check VAR2"},
        }
    }

    check = Check.load(d=d, name="base check")

    # Validate the root check
    assert check is not None
    assert isinstance(check, Check)
    assert check.name == "base check"

    # Validate the check tree
    flattened = check.flatten
    assert len(flattened) == 4

    # Validate class types
    assert flattened[0].__class__ == Check
    assert flattened[1].__class__ == Check
    assert flattened[2].__class__ == CheckDummy
    assert flattened[3].__class__ == CheckDummy

    # Validate check names
    assert flattened[0].name == "base check"
    assert flattened[1].name == "Environment"
    assert flattened[2].name == "Path"
    assert flattened[3].name == "Term"

    # Validate check values
    assert flattened[0].value is None
    assert flattened[1].value is None
    assert flattened[2].value == "VAR1"
    assert flattened[3].value == "VAR2"

    # Validate children entries
    assert flattened[0].children == [flattened[1]]
    assert flattened[1].children == [flattened[2], flattened[3]]
    assert flattened[2].children == []
    assert flattened[3].children == []


def test_check_import_modules():
    """Test the Check.import_modules method"""
    # Try a present module
    other_pytest = Check.import_modules("pytest")
    assert other_pytest == pytest

    # Try a missing module
    with pytest.raises(ImportError):
        Check.import_modules("nonexist-missing")


# noinspection GrazieInspection
def test_check_result_done():
    """Test Check.check and Result.done methods together."""

    # Create a check with a thread-locking sub-check
    sub = HangCheck(name="HangCheck")
    check = Check(name="root", children=[sub])

    # Create an executor to run the checks
    with ThreadPoolExecutor() as executor:
        result = check.check(executor=executor)

        # The 'sub' has the thread hanged and can't finish its check
        assert not result.done

        # Releasing the thread lock should allow 'sub' to finish its check
        sub.locked = False

        # The release might not be instantaneously after releasing the lock,
        # so we wait until the 'sub' check is done
        while not result.done:
            pass
        assert result.done


# noinspection GrazieInspection
@pytest.mark.parametrize("condition", ("all", "any"))
def test_check_result_passed(condition):
    """Test Check.check and Result.passed methods together (all condition."""

    # Create a check with a thread-locking sub-check
    sub1 = DefaultCheck(name="DefaultCheck1")
    sub2 = DefaultCheck(name="DefaultCheck2")
    check = Check(name="root", children=[sub1, sub2], condition=condition)

    # Create an executor to run the checks
    with ThreadPoolExecutor() as executor:
        result = check.check(executor=executor)

        # 1. Both sub1 and sub2 return results with status="passed" so the root
        #    check should pass under any 'condition'

        while not result.done:  # Wait for the thread to evaluate
            pass

        assert result.passed
        assert result.status == "passed"

        # 2. Changing one of the sub default statuses and rerunning the check
        #    should impact how the root check passes
        sub1.default_status = "failed"
        result = check.check(executor=executor)

        while not result.done:  # Wait for the thread to evaluate
            pass

        if condition == "any":
            assert result.passed  # Just one sub passing is good enough
            assert result.status == "passed"
        elif condition == "all":
            assert not result.passed  # One sub failed, so the root check fails
            assert result.status == "failed"

        # 3. Change both the sub default statuses and rerunning the check should
        #    impact how the root check passes
        sub2.default_status = "failed"
        result = check.check(executor=executor)

        while not result.done:  # Wait for the thread to evaluate
            pass

        assert not result.passed  # all children checks fail
        assert result.status == "failed"



"""
Checks the platform (Operating System) and its version
"""
import typing as t
import platform
import logging

from .base import Result, Executor
from .version import CheckVersion
from .utils import version_to_tuple
from ..config import Parameter

__all__ = ("CheckPlatform",)

logger = logging.getLogger(__name__)


class CheckPlatform(CheckVersion):
    """Check the availability and version of a python package"""

    # The message for checking python packages
    msg = Parameter(
        "CHECKPLATFORM.MSG",
        default="Check platform '{check.raw_value}'",
    )

    aliases = ("checkOS", "checkPlatform")

    def get_current_platform(self) -> str:
        """Retrieve the OS platform name"""
        uname = platform.uname()

        if uname.system in ("Darwin",):
            return "macOS"
        elif uname.system in ("Linux",):
            return "Linux"
        elif uname.system in ("Microsoft", "Windows"):
            return "Windows"
        else:
            raise NotImplementedError

    def get_current_version(self) -> t.Union[None, t.Tuple[int]]:
        """Retrieve the OS current version"""
        uname = platform.uname()

        # Get the current OS version
        current_platform = self.get_current_platform()

        # This code follows 'platform.platform()' to some extent
        if current_platform == "macOS":
            release = platform.mac_ver()[0]
            current_version = version_to_tuple(release) if release else None
        elif current_platform in ("Linux", "Windows"):
            current_version = version_to_tuple(uname.release) if uname.release else None
        else:
            current_version = None

        logger.debug(
            f"current_platform: {current_platform}, current_version: {current_version}"
        )
        return current_version

    def check(self, executor: t.Optional[Executor] = None, level: int = 0) -> Result:
        """Check whether the OS matches and the version"""
        msg = self.msg.format(check=self)

        # Get the OS name checked against
        name, op, version = self.value
        current_platform = self.get_current_platform()

        if current_platform.lower() != name.lower():
            # Failed check if the platform doesn't match this check
            return Result(msg=msg, status="wrong platform")

        # Check the version, as usual
        return super().check(level=level)



"""CLI utils"""
import typing as t
from pathlib import Path
import logging

__all__ = ("filepaths",)

logger = logging.getLogger(__name__)


def filepaths(string: str) -> t.List[Path]:
    """Given a string for a filepath or file glob, verifies that the path(s)
    exist converts to path object(s).

    Parameters
    ----------
    string
        The path string to verify and convert, which may include gobs

    Returns
    -------
    paths
        The paths for the existing files
    """
    # See if there's a glob pattern in the string
    if any(c in string for c in ("*", "?", "[", "]")):
        # Expand the glob from the current directory
        paths = list(Path(".").glob(string))
    else:
        paths = [Path(string)]

    # Check the paths
    existing_paths = []
    for path in paths:
        if path.is_file():
            existing_paths.append(path)

    return existing_paths



"""Test the CheckPythonPackage check"""
from geomancy.checks.python import CheckPythonPackage

try:
    import pip
except ImportError:
    pip = None


def test_check_python_package_exists():
    """Tests CheckPythonPackage checking for an existing and a missing
    package"""
    # Should exist
    check = CheckPythonPackage(name="Check pytest", value="pytest")
    assert check.check().passed

    # Should not exist
    check = CheckPythonPackage(name="Check missing", value="_miss_ing_")
    assert not check.check().passed


def test_check_python_package_version():
    """Tests CheckPythonPackage checking with version number"""
    # Should be greater than version 1.0
    check = CheckPythonPackage(name="Check pytest", value="pytest>=1.0")
    assert check.check().passed

    # Should be less than version 1000.
    check = CheckPythonPackage(name="Check pytest", value="pytest>=1000.0")
    assert not check.check().passed



"""
Abstract base class for checks
"""
import typing as t
import importlib
from concurrent.futures import Future, Executor
from types import ModuleType
from inspect import isabstract
from dataclasses import dataclass, field

from rich.table import Table
from rich.padding import Padding

from .utils import pop_first, all_subclasses
from ..config import Parameter
from ..environment import sub_env

__all__ = ("Check", "CheckException", "Result", "Executor")


class CheckException(Exception):
    """Exception raised when an error is encountered in the setup of a check."""


@dataclass(slots=True)
class Result:
    """A Check's result with awareness of concurrent.futures and rich
    functionality"""

    #: Result's status--e.g. 'passed', 'failed', 'not found'
    #: Only a 'passed' status is considered a passed result
    status: str = "pending"

    #: Result message used when displaying the result. This may include
    #: rich tags
    msg: str = ""

    #: The pass condition for children checks and their passed property values
    condition: t.Callable = all

    #: The flat list of Results or Futures from children checks
    children: t.List[t.Union["Result", t.Awaitable["Result"]]] = field(
        default_factory=list
    )

    @property
    def passed(self) -> bool:
        """Whether the check that generated this result passed.

        Notes
        -----
        This function checks whether children results have passed, whether this
        result has a non-pending status, and it updates "pending" status when
        the children are done
        """
        # Determine whether the children have finished
        done = True  # updated by loop below

        # Get the pass conditions for children. This list should only comprise
        # bools (True/False)
        children_passed = []
        for child in self.children:
            if isinstance(child, Result):
                # The child is a Result object
                child_status = child.status == "passed"
                children_passed.append(child_status)
            elif isinstance(child, Future) and child.done():
                # The child is a future with a result
                result = child.result()
                child_status = result.status == "passed"
                children_passed.append(child_status)
            else:
                # The child is a future that isn't done evaluating
                done = False
                children_passed.append(False)

        # Update the status from "pending" if this result is done (finished)
        if self.status.lower() == "pending" and done:
            self.status = "passed" if self.condition(children_passed) else "failed"

        # Only a 'passed' status is considered a passed result
        return self.status.lower() == "passed"

    @property
    def done(self) -> bool:
        """Whether the check that generated this result and children checks are
        done"""
        # Get the done status for children. This list should only comprise
        # bools (True/False)
        children_done = []
        for child in self.children:
            if isinstance(child, Result):
                # The child is a Result object
                child_done = child.done  # This function
                children_done.append(child_done)
            elif isinstance(child, Future) and child.done():
                # The child is a future with a result
                result = child.result()
                child_done = result.done  # This function
                children_done.append(child_done)
            else:
                children_done.append(False)

        if all(children_done) and self.status.lower() != "pending":
            # All children are done and this result has a non-pending status
            return True
        elif all(children_done) and self.status.lower() == "pending":
            # All the children are done and this result's pending. This result's
            # status can be updated/changed from 'pending'. The passed
            # property will do this
            _ = self.passed  # update status from "pending" to something else
            return True
        else:
            return False

    def rich_table(
        self, table: t.Optional[Table] = None, warning: bool = False, level: int = 0
    ) -> Table:
        """Generate a table with rich to display this result and child results

        Parameters
        ----------
        table
            The table object to add rows to. If it doesn't exist, a table will
            be created.
        warning
            If True, failed checks should be labeled as warnings instead of
            errors. This option is helpful when a parent check passes but a
            particular child check does not.
        level
            The level of the result in the nested tree

        Returns
        -------
        table
            The formatted table
        """
        # Create the table
        if table is None:
            table = Table(show_header=False, box=None, collapse_padding=True)
            table.add_column("Status")
            table.add_column("Value")

        # Format the checkbox string, status string and warning option for
        # children results
        if self.status.lower() == "pending":
            checkbox = "[ ]"
            status = f"[yellow]{self.status}[/yellow]"
        elif self.status.lower() == "passed":
            checkbox = "[[green]:heavy_check_mark:[/green]]"
            status = f"[green]{self.status}[/green]"
            warning |= True  # All children should be marked as warnings
        elif warning:
            checkbox = "[[yellow]![/yellow]]"
            status = f"[yellow]{self.status}[/yellow]"
        else:
            checkbox = "[[red]x[/red]]"
            status = f"[red]{self.status}[/red]"

        # 1. Add a row for self
        table.add_row(checkbox, Padding(f"{self.msg}...{status}", (0, 0, 0, 2 * level)))

        # 2. Add a row for each child
        for child in self.children:
            if isinstance(child, Result):
                # The child is a Result object
                child.rich_table(table=table, warning=warning, level=level + 1)
            elif isinstance(child, Future) and child.done():
                # The child is a future with a result
                result = child.result()
                result.rich_table(table=table, warning=warning, level=level + 1)

        return table


class Check:
    """Check base class and tree structure.

    .. versionchanged:: 1.0.0
        Rewrite and implemented threading using concurrent.futures.

    .. versionchanged:: 0.9.3
        Switch to :meth:`environment.sub_env` for value substitutions, which
        require a '$' character  and allow different expansion rules like
        defaults, errors and replacements.
    """

    #: The name for the check
    name: str

    #: Unprocessed value for the check
    raw_value: str

    #: Description of the check
    desc: str = ""

    #: The default message to include in results
    msg: str = "{check.name}"

    #: A list of children checks
    children: t.List["Check"]

    #: The condition function to use to evaluate whether children checks have passed
    condition: t.Callable = all  # default all must pass

    #: Alternative parameter names (__init__ kwarg names) used to specify the condition
    condition_aliases = ("condition", "subchecks")  # other names for variable

    #: Substitute environment variables in check values
    env_substitute: bool

    #: The default value for env_substitute
    env_substitute_default = Parameter("CHECK.ENV_SUBSTITUTE_DEFAULT", default=True)

    #: Alternative parameter names (__init__ kwarg names) for env_substitute
    env_substitute_aliases = ("env_substitute", "substitute")

    #: Default message and style of h1 headers
    h1_style = Parameter(
        "CHECK.H1_MSG", "[dodger_blue1][bold]{self.name}[/bold][/dodger_blue1]"
    )

    #: Default message and style of h2 headers
    h2_style = Parameter("CHECK.H2_MSG", "[bold]{self.name}[/bold]")

    #: Default message and style of h3 headers
    h3_style = Parameter("CHECK.H3_MSG", "[bold]{self.name}[/bold]")

    #: Default message and style of h4 headers
    h4_style = Parameter("CHECK.H4_MSG", "[bold]{self.name}[/bold]")

    #: Default message and style of h5 headers
    h5_style = Parameter("CHECK.H5_MSG", "[bold]{self.name}[/bold]")

    #: Default message and style of h6 headers (and lower)
    h6_style = Parameter("CHECK.H6_MSG", "[bold]{self.name}[/bold]")

    #: The import_module() exception message to use if a module is missing
    #: (see the :meth:`import_modules`)
    import_error_msg = "Missing dependency '{exception}'"

    #: Maximum recursion depth of the load function
    max_level = Parameter("CHECK.MAX_LEVEL", default=15)

    #: Whether this check class is available for use. Check subclasses may
    #: require additional dependencies, which might change this flag.
    #: (see :meth:`types`)
    available: bool = True

    #: Alternative names for the class (used by :meth:`~geomancy.checks.Check.types`)
    aliases: t.Optional[t.Tuple[str, ...]] = None

    def __init__(
        self,
        name: str,
        value: t.Optional[str] = None,
        desc: str = "",
        children: t.Optional[list["Check"]] = None,
        **kwargs,
    ):
        # Set attributes
        self.name = name
        self.value = value
        self.desc = desc
        self.children = list(children) if children is not None else []

        # Parse kwargs, which may use different aliases
        condition = pop_first(kwargs, *self.condition_aliases, default=None)
        self.env_substitute = pop_first(
            kwargs, *self.env_substitute_aliases, default=self.env_substitute_default
        )

        # Make sure the condition values are allowed
        if condition is None:
            pass
        elif condition.lower() == "all":
            self.condition = all
        elif condition.lower() == "any":
            self.condition = any
        else:
            raise CheckException(
                f"The condition '{condition}' should be either 'all' or 'any'."
            )

        # Check attributes
        if not all(isinstance(child, Check) for child in self.children):
            msg = "All children should be instances of Check"
            raise CheckException(msg)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"

    def __len__(self):
        return len(self.children)

    @property
    def value(self) -> t.Any:
        """Check's value with optional environment substitution"""
        if self.env_substitute and self.raw_value is not None:
            subbed = sub_env(self.raw_value)
            return subbed if subbed is not None else self.raw_value
        else:
            return self.raw_value

    @value.setter
    def value(self, v):
        self.raw_value = str(v) if v is not None else None

    @property
    def flatten(self) -> t.List["Check"]:
        """Return a flattened list of this check (first item) and children
        checks"""
        flattened = [self]
        for child in self.children:
            flattened += child.flatten
        return flattened

    @property
    def count(self) -> int:
        """The number of children, sub-children, etc, including self"""
        # flatten returns this check, so the count is subtracted by 1 to only
        # count sub-checks
        return len(self.flatten)

    @staticmethod
    def types() -> t.Dict[str, t.Type]:
        """The available types of Check classes, including aliases.

        Returns
        -------
        types_dict
            A dict containing the Check class or subclass name (key) and the
            class or subclass as values. The key-value pairs are also populated
            with alias names for Checks
        """
        # Retrieve the base class (Check) and children classes
        cls_types = [Check] + list(all_subclasses(Check))

        # Create a dict with the class name and type as key-value pairs
        d = dict()
        for cls_type in cls_types:
            # Skip abstract classes, which can't be instantiated
            if isabstract(cls_type):
                continue

            # Make sure class name isn't already in the dict
            assert (
                cls_type.__name__ not in d
            ), f"class {cls_type.__name__} already registered class"

            # Add the class name directly
            d[cls_type.__name__] = cls_type

            # Add class name aliases
            aliases = cls_type.aliases if cls_type.aliases is not None else ()
            for alias in aliases:
                # Aliases should not create name collisions
                assert alias not in d, f"Alias '{alias}' already matches '{d[alias]}'."

                d[alias] = cls_type
        return d

    @classmethod
    def load(
        cls, d: dict, name: str, level: int = 1, max_level: t.Optional[int] = None
    ) -> t.Union["Check", None]:
        """Load checks from a dict.

        Parameters
        ----------
        d
            The dict with checks to load
        name
            The name of the returned check instance
        level
            The current recursion depth of this load
        max_level
            The maximum recursion depth allowed

        Returns
        -------
        root_check
            The loaded root Check instance
        """
        # Check that the maximum recursion level hasn't been reached
        max_level = max_level if max_level is not None else cls.max_level

        if level >= max_level:
            msg = f"Parsing level {level} exceed the maximum level of {max_level}"
            raise NotImplementedError(msg)

        # Get a listing of the available Check types
        check_types = cls.types()

        # See if the dict has keys that reference Check types
        matching_keys = [k for k in d.keys() if k in check_types]

        if len(matching_keys) > 1:
            msg = f"More than 1 check type specified: {matching_keys}"
            raise NotImplementedError(msg)

        # Parse the check if a single check was given
        if len(matching_keys) == 1:
            # Get the check class
            check_type = matching_keys[0]
            matching_cls = check_types[check_type]

            # Get the value for the check
            value = d[check_type]

            # Get the other kwargs
            kwargs = {k: v for k, v in d.items() if k != check_type}

            # Create and return the check_type
            return matching_cls(name, value, **kwargs)

        # Otherwise, try parsing the children
        items = d.items()
        found_checks = []  # Values parsed into Check objects
        other_d = dict()  # All other values
        for key, value in items:
            if not isinstance(value, dict):
                other_d[key] = value
                continue

            return_value = cls.load(
                d=value, name=key, level=level + 1, max_level=max_level
            )

            # Replace the value withe Check instance, if it was parsed correctly
            # Otherwise, just place it in the parsed_dict.
            if isinstance(return_value, Check):
                found_checks.append(return_value)
            else:
                other_d[key] = value

        # No child checks found; nothing else to do
        if len(found_checks) == 0:
            return None

        # Create a check grouping, first, by parsing the other arguments
        return Check(name=name, children=found_checks, **other_d)

    @classmethod
    def import_modules(
        cls, *names: str
    ) -> t.Union[ModuleType, t.Tuple[ModuleType, ...]]:
        """Import and return modules given by the module name(s).

        This method is useful for loading modules for checks that require
        additional dependencies, which may not be installed--aws and the
        boto3 dependency, for example.

        Parameters
        ----------
        names
            The names of modules to return

        Returns
        -------
        modules
            The loaded module(s)

        Raises
        ------
        ImportError
            Raised if the modules couldn't be found, which can happen because
            the 'aws' extra dependency wasn't installed.
        """
        modules = []

        try:
            for name in names:
                module = importlib.import_module(name)
                modules.append(module)
        except ImportError as ie:
            msg = cls.import_error_msg.format(exception=ie)
            raise ImportError(msg)
        return modules[0] if len(modules) == 1 else tuple(modules)

    def check(self, executor: t.Optional[Executor] = None, level: int = 0) -> Result:
        """Performs this check and the children checks.

        Parameters
        ----------
        executor
            An object to submit or map calls asynchronously.
            See concurrent.futures.
        level
            The current depth level of the check tree

        Returns
        -------
        result
            The result of the check
        """
        assert (
            executor is not None
        ), "An executor must be specified to run children checks"

        # check children
        child_results = []
        for child in self.children:
            result = executor.submit(child.check, executor, level + 1)
            child_results.append(result)

        if level == 0:
            msg = self.h1_style.format(self=self)
        elif level == 1:
            msg = self.h2_style.format(self=self)
        elif level == 2:
            msg = self.h3_style.format(self=self)
        elif level == 3:
            msg = self.h4_style.format(self=self)
        elif level == 4:
            msg = self.h5_style.format(self=self)
        else:
            msg = self.h6_style.format(self=self)

        return Result(msg=msg, children=child_results, condition=self.condition)



"""Test the config module"""
from pathlib import Path

import pytest

from geomancy.config.config import ConfigException, Config, Parameter


@pytest.fixture
def reset_config():
    """Reset the Config singleton"""
    # noinspection PyProtectedMember
    if Config._instance is not None:
        Config._instance = None


def test_config_allowed_keys(reset_config):
    """Tests allowed key for Config entries"""
    config = Config()
    valid_keys = ("test1", "TEST1", "TEST_1", "_TEST1")
    invalid_keys = ("test 1", " test1", "test.1")

    for key in valid_keys:
        setattr(config, key, "value")

    for key in invalid_keys:
        with pytest.raises(ConfigException):
            setattr(config, key, "value")


def test_config_getset(reset_config):
    """Test the Config and Parameter get/set methods"""

    config = Config(value1="value1", value2=3)

    class A:
        attribute1 = Parameter("value1")
        attribute2 = Parameter("value2")

    a = A()

    # Try measuring the values directly
    assert config.value1 == "value1"
    assert config.value2 == 3
    assert a.attribute1 == "value1"
    assert a.attribute2 == 3

    # Try modifying the values on the config
    config.value1 = "new value1"
    config.value2 = 4

    # Try checking the values on the config object, the A class attribute and
    # the A instance attribute
    assert config.value1 == "new value1"
    assert config.value2 == 4
    assert A.attribute1 == "new value1"
    assert A.attribute2 == 4
    assert a.attribute1 == "new value1"
    assert a.attribute2 == 4

    # Try modifying the values on the class attributes
    a.attribute1 = "very new value1"
    a.attribute2 = 5

    # Try checking the new values on the config object, the A class attribute
    # and the A instance attribute
    assert config.value1 == "very new value1"
    assert config.value2 == 5
    assert A.attribute1 == "very new value1"
    assert A.attribute2 == 5
    assert a.attribute1 == "very new value1"
    assert a.attribute2 == 5


def test_config_default(reset_config):
    """Test the Config and Parameter with hard-coded default values."""

    # Try creating attributes with defaults
    class A:
        attribute1 = Parameter("A.attribute1", default="default value")
        attribute2 = Parameter("A.attribute2", default=3)

    # Check that these values were placed in the config
    config = Config()
    assert config.A.attribute1 == "default value"
    assert config.A.attribute2 == 3

    # Try changing the config values and see if those changes have been
    # reflected in the config
    config.A.attribute1 = "new default value"
    config.A.attribute2 = 4

    # The class values should be changed
    assert A.attribute1 == "new default value"
    assert A.attribute2 == 4


def test_config_nested(reset_config):
    """Test the Config and Parameter with nested attributes"""
    config = Config()

    # Try modifying the config
    config.general.value1 = "value 1"
    config.general.value2 = 2

    # Try accessing through class properties
    class A:
        attribute1 = Parameter("general.value1")
        attribute2 = Parameter("general.value2")

    a = A()

    # Check the nested dict in Config - level 1
    assert isinstance(config.__dict__, dict)
    assert len(config.__dict__) == 1

    # Check the nested dict in Config - level 2
    assert len(config.__dict__["general"]) == 2
    assert isinstance(config.__dict__["general"], Config)

    # Check the nested dict in Config - level 3
    assert len(config.__dict__["general"].__dict__) == 2
    assert config.__dict__["general"].__dict__["value1"] == "value 1"
    assert config.__dict__["general"].__dict__["value2"] == 2

    # Try measuring the values directly on the Config singleton and the
    # class/instance attributes
    assert config.general.value1 == "value 1"
    assert config.general.value2 == 2
    assert A.attribute1 == "value 1"
    assert A.attribute2 == 2
    assert a.attribute1 == "value 1"
    assert a.attribute2 == 2

    # Try modifying the values on the Config
    config.general.value1 = "new value 1"
    config.general.value2 = 3

    # Try measuring the values directly on the Config singleton and the
    # class/instance attributes
    assert config.general.value1 == "new value 1"
    assert config.general.value2 == 3
    assert A.attribute1 == "new value 1"
    assert A.attribute2 == 3
    assert a.attribute1 == "new value 1"
    assert a.attribute2 == 3

    # Try modifying the values on the instance
    a.attribute1 = "very new value 1"
    a.attribute2 = 4

    # Try measuring the values directly on the Config singleton and the
    # class/instance attributes
    assert config.general.value1 == "very new value 1"
    assert config.general.value2 == 4
    assert A.attribute1 == "very new value 1"
    assert A.attribute2 == 4
    assert a.attribute1 == "very new value 1"
    assert a.attribute2 == 4


def test_config_recursive_update(reset_config):
    """Test the recursive update"""
    values = {
        "section1": {"subsection1": 1, "subsection2": 2},
        "section2": 3,
        "section3": 4,
    }
    updates = {
        "section1": {"subsection2": -2},
        "section2": -3,
        "section4": {"subsection4": 5},
    }
    bad_updates = {"section1": 0}  # Replace a section with a value

    # Create and check the reference config
    config = Config.load(values)

    assert config.section1.subsection1 == 1
    assert config.section1.subsection2 == 2
    assert config.section2 == 3
    assert config.section3 == 4

    # Introduce updates and check new values
    config.update(updates)

    assert config.section1.subsection1 == 1
    assert config.section1.subsection2 == -2  # new value
    assert config.section2 == -3  # new value
    assert config.section3 == 4
    assert config.section4.subsection4 == 5  # new section

    # Try a bad update that eliminates a section
    with pytest.raises(ConfigException):
        config.update(bad_updates)

    assert config.section1.subsection1 == 1  # config remains


@pytest.mark.parametrize(
    "ext,save_file",
    (
        ("toml", False),
        ("toml", True),
        ("yaml", False),
        ("yaml", True),
    ),
)
def test_config_parsing_strings(reset_config, tmp_path, ext, save_file):
    """Test Config load/dump string methods"""
    # Setup a config
    config = Config()
    config.a = 1
    config.b = "string"
    config.c = True
    config.d.a = "subsection1"
    config.d.b = "subsection2"

    # Create a string from the default config
    dict_copy = dict(config.__dict__)
    assert len(dict_copy) == 4  # 'a', 'b', 'c'  and 'd'

    # Create a string from the config
    meth = getattr(config, f"dumps_{ext}")
    config_str = meth()
    assert config_str.strip() != ""

    # Optionally save the string to a file
    if save_file:
        filepath = (tmp_path / "test").with_stem("." + ext)
        filepath.write_text(config_str)
    else:
        filepath = None

    Config._instance = None  # reset singleton

    # Try reloading the string
    if save_file:
        meth = getattr(Config, f"load_{ext}")
        new_config = meth(filepath)
    else:
        meth = getattr(Config, f"loads_{ext}")
        new_config = meth(config_str)

    # Make sure the new dict matches the old
    assert len(dict_copy) == 4 and len(new_config.__dict__) == 4
    assert dict_copy == new_config.__dict__


def test_config_toml_parsing(reset_config):
    """Test Config with the parsing of TOML strings and files"""
    # Load the toml file
    filename = Path(__file__).parent / "config1.toml"
    config = Config.load_toml(filename)

    # Check the parsed config
    assert config.checkEnv.env_substitute
    assert config.checkEnv.msg == "A test message"
    assert config.checkEnv.value.nested == 1



"""The geo CLI entrypoint"""
import logging
import os
from pathlib import Path

import click
from click_default_group import DefaultGroup

from .check import check
from .run import run
from .config import config
from .. import get_version

logger = logging.getLogger(__name__)

description = (Path(__file__).parent / ".." / "__description__.txt").read_text().strip()


def print_version(context, parameter, value):
    """Print the version"""
    if not value or context.resilient_parsing:
        return
    click.echo(get_version())
    context.exit()


@click.group(
    cls=DefaultGroup, default="check", default_if_no_args=True, help=description
)
@click.option("--debug", "-d", is_flag=True, help="Enable debugging information")
@click.option(
    "--version",
    "-V",
    is_flag=True,
    is_eager=True,
    expose_value=False,
    callback=print_version,
    help="Show the version and exit.",
)
@click.option("--disable-color", is_flag=True, help="Disable terminal coloring")
@click.pass_context
def geo_cli(ctx, debug, disable_color):
    """The main entrypoint for the 'geo' CLI"""
    # Disable coloring, if specified
    if disable_color:
        os.environ["NO_COLOR"] = "TRUE"

    # Setup logging
    logging.basicConfig(
        level=logging.DEBUG if debug else None,
        format="%(levelname)s:%(name)s: %(message)s",
    )


# Add sub-commands
geo_cli.add_command(check)
geo_cli.add_command(run)
geo_cli.add_command(config)



"""Checks for AWS resources"""
import typing as t
import logging

from ..base import Check, Result, Executor
from ..utils import pop_first
from ...config import Parameter

__all__ = ("CheckAWSS3",)

logger = logging.getLogger(__name__)

#: The error message to show when AWS modules can't be imported
import_error_msg = (
    "The 'aws' dependency is not installed: {exception}. "
    "Please reinstall with the '[aws]' or '[all]' extra install "
    '`python -m pip install "geomancy[aws]"` or '
    '`python -m pip install "geomancy[all]"`'
)


class CheckAWSS3BucketAccess(Check):
    """Check AWS S3 bucket availability"""

    msg = Parameter(
        "CHECKAWSS3BUCKETACCESS.MSG",
        default="Check AWS S3 bucket access '{check.value}'...",
    )

    import_error_msg = import_error_msg

    def check(self, executor: t.Optional[Executor] = None, level: int = 0) -> Result:
        """Check the availability and access to S3 Bucket"""
        msg = self.msg.format(check=self)

        # Get the needed modules, bucket name and boto3 client
        boto3, exceptions = self.import_modules("boto3", "botocore.exceptions")
        bucket_name = self.value.strip()
        s3 = boto3.client("s3")

        # Retrieve information on the bucket
        try:
            response = s3.head_bucket(Bucket=bucket_name)

        except exceptions.NoCredentialsError as e:
            # Unable to authenticate the client
            return Result(msg=msg, status="Unable to locate credentials")

        except exceptions.ClientError as e:
            response = e.response

            # Retrieve error information from the response
            error = response["Error"] if "Error" in response else None
            error_code = error["Code"] if error and "Code" in error else ""
            error_msg = error["Message"] if error and "Message" in error else ""

            logger.debug(f"s3.head_bucket error: {error}")

            if error_msg in ("Not Found",) and error_code == "404":
                # Couldn't find the bucket
                return Result(msg=msg, status="not found")
            elif error_msg in ("Forbidden",) and error_code == "403":
                # Do no have access to an existing bucket
                return Result(msg=msg, status="access forbidden")
            else:
                return Result(msg=msg, status="unknown reason")

        except exceptions.ParamValidationError as e:
            # The bucket name failed validation
            report = e.kwargs["report"] if "report" in e.kwargs else None

            if isinstance(report, str):
                return Result(msg=msg, status=report)
            else:
                return Result(msg=msg, status="invalid bucket name")

        # Parse the response
        metadata = (
            response["ResponseMetadata"] if "ResponseMetadata" in response else None
        )
        return_code = (
            metadata["HTTPStatusCode"]
            if metadata and "HTTPStatusCode" in metadata
            else None
        )

        if return_code == 200:
            # Successfully probed S3 bucket
            return Result(msg=msg, status="passed")
        else:
            # It failed, and I don't know why
            return Result(msg=msg, status="unknown reason")


class CheckAWSS3BucketPrivate(Check):
    """Check AWS S3 buck availability"""

    msg = Parameter(
        "CHECKAWSS3BUCKETPRIVATE.MSG",
        default="Check AWS S3 bucket private '{check.value}'...",
    )

    import_error_msg = import_error_msg

    def check(self, executor: t.Optional[Executor] = None, level: int = 0) -> Result:
        """Check the availability and access to S3 Bucket.

        See: https://stackoverflow.com/a/59002759/9099988
        """
        msg = self.msg.format(check=self)

        # Get the needed modules, bucket name and boto3 client
        boto3, exceptions = self.import_modules("boto3", "botocore.exceptions")
        bucket_name = self.value.strip()
        s3 = boto3.client("s3")

        # 1. Check the PublicAccessBlock
        try:
            response = s3.get_public_access_block(Bucket=bucket_name)
        except (exceptions.BotoCoreError, exceptions.ClientError):
            return Result(msg=msg, status="Couldn't access PublicAccessBlock")

        section = response.get("PublicAccessBlockConfiguration", {})
        block_public_policy = section.get("BlockPublicPolicy")
        block_public_acls = section.get("BlockPublicAcls")

        if block_public_policy and block_public_acls:
            return Result(msg=msg, status="passed")

        logger.debug(f"Bucket '{bucket_name}' does not block all public access")

        # 2. Check the Bucket policy to see if it allows public access
        try:
            response = s3.get_bucket_policy_status(Bucket=bucket_name)
        except exceptions.ClientError:
            # No policy found
            response = {}

        # See if a policy was set and whether it allows public access
        section = response.get("PolicyStatus", {})
        public_policy = section.get("IsPublic", False)

        logger.debug(f"Bucket '{bucket_name}' public policy is: {public_policy}")

        # 3. Check the Bucket ACL to see if it allows public access
        try:
            response = s3.get_bucket_acl(Bucket=bucket_name)
        except exceptions.ClientError:
            # No policy found
            response = {}

        # See if ACL grants access
        section = response.get("Grants", {})
        grantee_types = [item.get("Grantee", {}).get("Type", "") for item in section]
        public_acl = any(grantee_type == "Group" for grantee_type in grantee_types)

        logger.debug(f"Bucket '{bucket_name}' ACL to groups is: {public_acl}")

        # Summary
        # If the policy is false and the acl doesn't have a grantee type of 'group',
        # then it should be private
        if not public_policy and not public_acl:
            return Result(msg=msg, status="passed")

        # At this stage, all the checks for private have failed, and it appears
        # that the s3 bucket is publicly accessible
        return Result(msg=msg, status="publicly accessible")


class CheckAWSS3(Check):
    """Check AWS the availability and permissions of S3 buckets.

    Notes
    -----
    The child checks are meant to verify the accessibility of the bucket as well
    as optionally test for the best security practices`. Current, these include:

    - Default private access (CheckAWSS3BucketPrivate)

    See: https://docs.aws.amazon.com/AmazonS3/latest/userguide/security-best-practices.html
    """

    #: Whether to check that the bucket is private
    private: bool = True

    #: Alternative parameter names for private
    private_aliases = ("private",)

    msg = Parameter(
        "CHECKAWSS3.MSG",
        default="Check AWS S3 bucket '{check.value}'...",
    )

    aliases = ("checkAWSS3", "checkAwsS3", "CheckAwsS3", "checkS3", "CheckS3")

    def __init__(self, *args, **kwargs):
        # Set up keyword arguments
        self.private = pop_first(kwargs, *self.private_aliases, default=self.private)

        # Set up the rest of the class
        super().__init__(*args, **kwargs)

        # Replace children with bucket sub-checls
        self.children.clear()

        # Bucket accessibility check
        self.children.append(
            CheckAWSS3BucketAccess(name=f"{self.name}Access", value=self.value)
        )

        # Bucket public access
        if self.private:
            self.children.append(
                CheckAWSS3BucketPrivate(name=f"{self.name}Private", value=self.value)
            )

    def check(self, executor: t.Optional[Executor] = None, level: int = 0) -> Result:
        """Run sub-checks in sequence, rather the Check base default.

        This is done because the results of latter checks depend on earlier checks.
        """
        # check children
        child_results = []
        for child in self.children:
            result = child.check(executor=executor, level=level + 1)
            child_results.append(result)
            if result.status != "passed":
                break

        return Result(
            msg=f"[bold]{self.name}[/bold]",
            children=child_results,
            condition=self.condition,
        )



"""Tests for the platform and version"""
from geomancy.checks import CheckPlatform


def test_check_platform():
    """Test the CheckPlatform check"""
    # Minimum version for Python 3.11 is macOS 10.9
    check_macos = CheckPlatform(name="checkMacOS", value="macOS >= 10.9")

    # Linux 4.0 was released in 2015
    check_linux = CheckPlatform(name="checkLinux", value="Linux >= 4.0")

    # Windows 10 was released in 2015
    check_windows = CheckPlatform(name="checkWindows", value="Windows >= 10")

    assert any(
        check.check().passed for check in (check_macos, check_linux, check_windows)
    )



"""Test the main CLI entrypoint"""
import typing as t
from pathlib import Path
import os

from click.testing import CliRunner
import pytest

from geomancy.entrypoints import geo_cli
from geomancy.config import Config


def get_checks_files():
    """Return a list of checks files to check in the root examples/"""
    checks_files = []

    # examples/
    for ext in ("toml", "yaml"):
        checks_files += list(Path("examples").glob(f"*.{ext}"))

    return checks_files


@pytest.fixture
def run() -> t.Callable:
    """Run the CLI with the given option, check for the expected exit
    code and return the output"""

    def runcmd(args, expected_code: int = 0):
        args = [args] if isinstance(args, str) else args

        runner = CliRunner()
        result = runner.invoke(geo_cli, args)
        if result.exit_code != expected_code:
            print(result.output)
        assert result.exit_code == expected_code
        return result

    runcmd.__doc__ = run.__doc__
    return runcmd


@pytest.fixture
def config() -> Config:
    return Config()


@pytest.mark.parametrize("options", ("--help",))
def test_cli_help(run, options):
    """Test the --help message"""
    result = run(options)
    assert "Show this message and exit" in result.output


@pytest.mark.parametrize("options", get_checks_files())
def test_cli_check(run, options):
    """Test CLI with all the checks files individually in the examples directory"""
    result = run(str(options))

    # Check, for example, that environment variables were checked
    assert any(
        msg in result.output for msg in ("Check environment variable", "Check path")
    )


@pytest.mark.parametrize("options", ("examples/geomancy.*",))
def test_cli_check_glob(run, options):
    """Test the CLI with glob patterns"""
    result = run(str(options))

    # Check, for example, that environment variables were checked
    assert any(
        msg in result.output for msg in ("Check environment variable", "Check path")
    )


@pytest.mark.parametrize(
    "checks_file,fixture",
    (("examples/aws/geomancy.yaml", "examples/aws/fixtures.yaml"),),
)
def test_cli_check_fixtures(run, checks_file, fixture):
    """Test the CLI with checks that require network access and therefore use
    fixtures."""
    # Run the command and test that an 0 error code was received
    result = run(("--fixture", fixture, checks_file))


@pytest.mark.parametrize("flag", ("", "--toml", "--yaml"))
def test_cli_config(run, flag):
    """Test the --config option"""
    if flag:
        result = run(("config", flag))
    else:
        result = run(("config",))
    assert "config" in result.output  # config output in TOML/yaml format


@pytest.mark.parametrize("flag", ("-e", "--env"))
def test_cli_handle_env(run, flag, test_env_file):
    """Test the handling of -e/--env and --overwrite flags for loading
    environment variables.

    See ./conftest.py for details on the 'test_env_file' fixture
    """
    # Get a single checks file to test against
    checks_file = get_checks_files()[0]

    # Set up the env files
    env_filepath = test_env_file["filepath"]
    variables = test_env_file["variables"]

    # running '--overwrite' without '-e/--env' gives an error
    result = run("--overwrite", expected_code=2)

    with pytest.MonkeyPatch.context() as mp:
        # Reset env variables
        for name in variables.keys():
            mp.delenv(name, raising=False)

        # The variables are in the environment (os.environ)
        for name, value in variables.items():
            assert name not in os.environ

        # running "-e/--env" should load environment variables, which are logged
        # in debug mode
        options = ("-d", flag, env_filepath, "checks", str(checks_file))
        result = run(options)

        # The variables are loaded in the current process
        for name, value in variables.items():
            assert os.environ[name] == value


def test_cli_run(run, test_env_file):
    """Test the 'run' subcommand and the handle_env function.

    See ./conftest.py for details on the 'test_env_file' fixture.
    """
    # Set up the env files
    filepath = test_env_file["filepath"]
    variables = test_env_file["variables"]

    with pytest.MonkeyPatch.context() as mp:
        # Reset env variables
        for name in variables.keys():
            mp.delenv(name, raising=False)

        # The variables are in the environment (os.environ)
        for name, value in variables.items():
            assert name not in os.environ

        # Run the command with the 'run' subcommand
        result = run(("run", "-e", filepath, "echo", "here!"))

        # The variables are loaded in the current process
        for name, value in variables.items():
            assert os.environ[name] == value



"""Test env functions, regexes and utilities"""
import os

import pytest

from geomancy.environment import sub_env, parse_env, load_env


def test_sub_env():
    """Test environment variable substitutions with sub_env"""
    with pytest.MonkeyPatch.context() as mp:
        mp.setenv("VAR1", "variable1")
        mp.setenv("VAR2", "variable2")
        mp.delenv("MISSING", raising=False)

        # Test with braces and with $
        assert sub_env("${VAR1}") == "variable1"
        assert sub_env("${VAR2}") == "variable2"
        assert sub_env("${MISSING}") == ''

        # Test without braces and with $
        assert sub_env("$VAR1") == "variable1"
        assert sub_env("$VAR2") == "variable2"
        assert sub_env("$MISSING") == ''


def test_parse_env_docker_rules():
    """Test the parse_env function rules compared to docker dotenv rules.

    The layout and order of these tests deliberately matches the following:
    https://docs.docker.com/compose/environment-variables/env-file/#syntax
    """
    p = parse_env

    # Lines beginning with # are processed as comments and ignored
    assert p("# VAR=VAL") == {}
    assert p(" # VAR=VAL") == {}

    # Blank lines are ignored
    assert p("\n") == {}

    # Each line represents a key-value pair. Values can optionally be quoted
    assert p("VAR=VAL") == {"VAR": "VAL"}
    assert p('VAR="VAL"') == {"VAR": "VAL"}
    assert p("VAR='VAL'") == {"VAR": "VAL"}

    # Inline comments for unquoted values must be preceded with a space
    assert p("VAR=VAL # comment") == {"VAR": "VAL"}
    assert p("VAR=VAL# not a comment") == {"VAR": "VAL# not a comment"}

    # Inline comments for quoted values must follow the closing quote
    assert p('VAR="VAL # not a comment"') == {"VAR": "VAL # not a comment"}
    assert p('VAR="VAL" # comment') == {"VAR": "VAL"}

    with pytest.MonkeyPatch.context() as mp:
        mp.setenv("OTHER", "NEWVALUE")

        # Single-quoted (') values are used literally.
        assert p("VAR='$OTHER'") == {"VAR": "$OTHER"}
        assert p("VAR='${OTHER}'") == {"VAR": "${OTHER}"}

        # Double-quote (") values allow substitution
        assert p(r'VAR="My ${OTHER}."') == {"VAR": "My NEWVALUE."}

    # Quotes can be escaped with a backslash: \' or \"
    assert p(r"VAR='Let\'s go!'") == {"VAR": r"Let's go!"}
    assert p(r'VAR="{\"hello\": \"json\"}"') == {"VAR": r'{"hello": "json"}'}

    # Common shell escape sequences including \n, \r, \t, and \\ are supported
    # in double-quoted values.
    assert p(r'VAR="some\tvalue"') == {"VAR": "some\tvalue"}
    assert p(r"VAR='some\tvalue'") == {"VAR": r"some\tvalue"}
    assert p(r"VAR=some\tvalue") == {"VAR": r"some\tvalue"}


def test_parse_env_docker_parameter_expansion():
    """Test the parse_env_str function parameter expansion compared to docker
    dotenv parameter expansion.

    The layout and order of these tests deliberately matches the following:
    https://docs.docker.com/compose/environment-variables/env-file/#parameter-expansion
    """
    p = parse_env

    with pytest.MonkeyPatch.context() as mp:
        mp.setenv("OTHER_VAL", "new value")

        # Direct substitution (braced)
        assert p(r"VAR=${OTHER_VAL}") == {"VAR": "new value"}  # (un-quoted)
        assert p(r'VAR="${OTHER_VAL}"') == {"VAR": "new value"}  # (double quoted)
        assert p(r"VAR='${OTHER_VAL}'") == {"VAR": "${OTHER_VAL}"}  # (literal)

        # Direct substitution (unbraced)
        assert p(r"VAR=$OTHER_VAL") == {"VAR": "new value"}  # (un-quoted)
        assert p(r'VAR="$OTHER_VAL"') == {"VAR": "new value"}  # (double quoted)
        assert p(r"VAR='$OTHER_VAL'") == {"VAR": "$OTHER_VAL"}  # (literal)

        # Default value
        assert p(r"VAR=${MISSING:-my default value}") == {"VAR": "my default value"}
        assert p(r"VAR=${MISSING-my default value}") == {"VAR": "my default value"}
        assert p(r"VAR=$MISSING:-default") == {"VAR": "default"}
        assert p(r"VAR=$MISSING-default") == {"VAR": "default"}

        # Error value
        for error_value in (
            r"VAR=${ERROR:?error raised}",
            r"VAR=${ERROR?error raised}",
            r"VAR=$ERROR:?error",
            r"VAR=$ERROR?error",
        ):
            with pytest.raises(EnvironmentError) as e:
                p(error_value)

        # Replace value
        assert p(r"VAR=${OTHER_VAL:+replaced value}") == {"VAR": "replaced value"}
        assert p(r"VAR=${OTHER_VAL+replaced value}") == {"VAR": "replaced value"}
        assert p(r"VAR=$OTHER_VAL:+replaced") == {"VAR": "replaced"}
        assert p(r"VAR=$OTHER_VAL+replaced") == {"VAR": "replaced"}

        assert p(r"VAR=${MISSING:+replaced value}") == {"VAR": ""}
        assert p(r"VAR=${MISSING+replaced value}") == {"VAR": ""}
        assert p(r"VAR=$MISSING:+replaced") == {"VAR": ""}
        assert p(r"VAR=$MISSING+replaced") == {"VAR": ""}


def test_load_env(test_env_file):
    """Test the load_env function using test.env.

    See ../conftest.py for details on the 'test_env_file' fixture
    """
    with pytest.MonkeyPatch.context() as mp:
        filepath = test_env_file["filepath"]
        variables = test_env_file["variables"]

        # Clear the environment variables from 'test.env'
        for name in variables.keys():
            mp.delenv(name, raising=False)

        # Load the environment variables from 'test.env'
        env_vars = load_env(filepath=filepath, overwrite=False)

        for name, value in variables.items():
            assert os.environ[name] == value
        assert len(env_vars) == len(variables)

        # The environment variables are already loaded, so trying so again
        # will give a loaded count = 0 -- i.e. no variables loaded
        env_vars = load_env(filepath=filepath, overwrite=False)
        assert len(env_vars) == 0

        # Unless we put the overwrite flag, in which case the 5 variables will
        # be overwritten
        env_vars = load_env(filepath=filepath, overwrite=True)
        assert len(env_vars) == len(variables)



from .base import Check, CheckException, Result
from .env import CheckEnv
from .exec import CheckExec
from .path import CheckPath
from .platform import CheckPlatform
from .python import CheckPythonPackage

from .aws import CheckAWSS3

__all__ = (
    Check,
    CheckException,
    Result,
    CheckEnv,
    CheckExec,
    CheckPath,
    CheckPlatform,
    CheckPythonPackage,
    CheckAWSS3,
)



"""
Checks for paths
"""
import typing as t
from pathlib import Path

from .base import Check, Result, CheckException, Executor
from ..config import Parameter

__all__ = ("CheckPath",)


class CheckPath(Check):
    """Check paths for valid files and directories"""

    #: (Optional) the type of path expected
    type: t.Optional[str] = None

    #: The valid values of path types
    type_options = (None, "dir", "file")

    msg = Parameter("CHECKPATH.MSG", default="Check path '{check.value}'")

    aliases = ("checkPath",)

    def __init__(self, *args, type: t.Optional[str] = None, **kwargs):
        super().__init__(*args, **kwargs)
        if type not in self.type_options:
            raise CheckException(
                f"Path type '{type}' must be one of: "
                f"{tuple(opt for opt in self.type_options if opt is not None)}"
            )
        self.type = type

    def check(self, executor: t.Optional[Executor] = None, level: int = 0) -> Result:
        """Check paths"""
        value = self.value
        path = Path(value)

        if not path.exists():
            status = "missing"
        elif self.type == "dir" and not path.is_dir():
            status = "not dir"
        elif self.type == "file" and not path.is_file():
            status = "not file"
        else:
            status = "passed"

        msg = self.msg.format(check=self, status=status)
        return Result(msg=msg, status=status)



"""
The 'check' subcommand
"""
import typing as t
import logging
import tomllib
import time
from concurrent.futures import ThreadPoolExecutor
from contextlib import ExitStack

import click
from rich.live import Live
import yaml
import vcr

from .environment import env_options
from .utils import filepaths
from ..checks import Check
from ..config import Config

__all__ = ("check",)

# Setup logger and configuration
logger = logging.getLogger(__name__)


# Exception classes
class MissingChecks(click.ClickException):
    """No checks were found in the checks files"""


# Config options
config = Config()
config.CLI.CHECKS_PATHS = [  # Default paths for checks files
    "pyproject.toml",
    ".geomancy.??ml",
    "geomancy.??ml",
    "geomancy.yml",
    ".geomancy.yml",
]
config.CLI.TOML_EXTS = [".toml"]  # Default file extensions for TOML files
config.CLI.YAML_EXTS = [".yml", ".yaml"]  # Default file extensions for YAML files
config.CLI.VCR.RECORD_MODE = "ONCE"  # record fixtures if they don't exist
config.CLI.VCR.ADDITIONAL_FILTER_HEADERS = []


def validate_checks_files(
    ctx: click.Context, param: click.Parameter, values: t.Tuple[str]
):
    """Validate the checks files arguments and convert to valid paths"""
    # Convert filepath strings into Path objects. Use default locations if
    # no checks_files were specified (i.e. it is an empty list)
    existing_files = []
    for path in values or config.CLI.CHECKS_PATHS:
        existing_files += filepaths(path)

    # Nothing to do if no checks files were found
    if len(existing_files) == 0:
        raise click.MissingParameter(
            "Could not find a checks file.", ctx=ctx, param=param
        )
    logging.debug(f"Checking the following files: {existing_files}")
    return existing_files


# Setup 'check' command
@click.command
@env_options
@click.option(
    "-f",
    "--fixture",
    required=False,
    type=click.Path(),
    help="Fixture to replace network requests",
)
@click.argument("checks_files", nargs=-1, type=str, callback=validate_checks_files)
def check(checks_files, env, fixture):
    """Run checks"""
    logger.debug(f"check_files={checks_files}, env={env}, fixture={fixture}")

    # Convert the checks_files into checks
    checks = []
    for checks_file in checks_files:
        # Parse the file by filetype
        if checks_file.suffix in config.CLI.TOML_EXTS:
            with open(checks_file, "rb") as f:
                d = tomllib.load(f)

        elif checks_file.suffix in config.CLI.YAML_EXTS:
            with open(checks_file, "r") as f:
                d = yaml.load(f, Loader=yaml.SafeLoader)

        else:
            continue

        # pyproject.toml files have their items placed under the [tool.geomancy]
        # section
        if checks_file.name == "pyproject.toml":
            d = d.get("tool", dict()).get("geomancy", dict())

        # Load config section, if available
        for config_name in config.section_aliases:
            config_section = d.pop(config_name, None)
            if isinstance(config_section, dict):
                config.update(config_section)

        # Load the rest into a root CheckBase
        check = Check.load(d, name=str(checks_file))
        if check is not None:
            checks.append(check)

    # Create a root check, if there are a lot of checks
    if len(checks) > 1:
        check = Check(name=f"Checking {len(checks)} files", children=checks)
    elif len(checks) == 1:
        check = checks[0]
    else:
        plural = True if len(checks_files) > 1 else False
        raise MissingChecks(
            f"No checks were found in the file{'s' if plural else ''}: "
            f"{', '.join(map(str, checks_files))}."
        )

    with ExitStack() as stack:
        # Set up the context managers
        executor = stack.enter_context(ThreadPoolExecutor())  # concurrent.futures
        live = stack.enter_context(Live(refresh_per_second=4))  # rich live display

        if fixture is not None:
            # Request header items to remove before saving
            filter_headers = {
                "Authorization",
                "User-Agent",
                "X-Amz-Content-SHA256",
                "X-Amz-Date",
                "amz-sdk-invocation-id",
                "amz-sdk-request",
            }
            filter_headers.update(config.CLI.VCR.ADDITIONAL_FILTER_HEADERS)
            kwargs = {
                "record_mode": config.CLI.VCR.RECORD_MODE,
                "filter_headers": list(filter_headers),
            }

            stack.enter_context(vcr.use_cassette(fixture, **kwargs))  # vcr fixtures

        # Run the checks, display the results to the terminal
        result = check.check(executor=executor)

        # Update the display until the checks are done
        while not result.done:
            time.sleep(0.5)
            live.update(result.rich_table())

        # Print the final table
        live.update(result.rich_table())

    if not result.passed:
        exit(1)

    return result.passed



"""
Test the CheckExec class
"""
from geomancy.checks.exec import CheckExec


def test_check_exec_get_current_version_no_pip():
    """Test the CheckExec get_current_version method"""
    check = CheckExec(name="python", value="python")

    version = check.get_current_version()
    assert isinstance(version, tuple)
    assert version[0] >= 2  # python 2 or later


def test_check_exec_exists():
    """Tests CheckExec checking for an existing and a missing command"""
    # Should exist
    check = CheckExec(name="Check Python", value="python3")
    assert check.check().passed

    # Should not exist
    check = CheckExec(name="Check Python", value="_miss_ing_")
    assert not check.check().passed


def test_check_exec_version():
    """Tests CheckExec checking with version number"""
    # Should be greater than version 2.0
    check = CheckExec(name="Check Python", value="python>=2.0")
    assert check.check().passed

    # Should be less than version 1000.
    check = CheckExec(name="Check Python", value="python>=1000.0")
    assert not check.check().passed



"""
Test the path check class
"""
from pathlib import Path

import pytest

from geomancy.checks import CheckPath, CheckException


def test_check_path_missing():
    """Test CheckPath with a missing path"""
    missing_path = "./.missing.miss"
    assert not Path(missing_path).exists()

    check = CheckPath("missing", missing_path)
    assert check.name == "missing"
    assert check.value == missing_path
    assert not check.check().passed


@pytest.mark.parametrize("path_type", (None, "file", "dir", "nonsense"))
def test_check_path_file_type(tmp_path, path_type):
    """Test CheckPath with different file types"""
    # Create the test files and dirs
    tmp_file = tmp_path / "exists.txt"
    tmp_file.touch()

    tmp_dir = tmp_path / "sub"
    tmp_dir.mkdir()

    # Create the check with an existing file
    if path_type == "nonsense":
        with pytest.raises(CheckException):
            CheckPath(name="PathCheck", value=tmp_file, type=path_type)
        return None
    else:
        check = CheckPath(name="PathCheck", value=tmp_file, type=path_type)

    if path_type in (None, "file"):
        assert check.check().passed
    else:
        assert not check.check().passed

    # Switch to the directory, and the check conditions change
    check.value = tmp_dir

    if path_type in (None, "dir"):
        assert check.check().passed
    else:
        assert not check.check().passed


def test_check_path_env_substitution(tmp_path):
    """Test CheckPath with environmental subsitution"""
    # Create the test files and dirs
    ENV = "dev"
    tmp_file = tmp_path / ENV / "exists.txt"
    tmp_file.parent.mkdir()
    tmp_file.touch()

    path = tmp_path / "$ENV" / "exists.txt"
    assert "$ENV" in str(path)  # not substituted yet

    with pytest.MonkeyPatch.context() as mp:
        # Make sure the ENV variable hasn't been created yet
        mp.delenv("ENV", raising=False)

        # The variable doesn't exist so the check should fail
        check = CheckPath(name="PathEnvVariable", value=path)
        assert not check.check().passed

        # Set the ENV variable, and it should now work
        mp.setenv("ENV", ENV)
        assert check.check().passed



"""
Checks for python environment and packages
"""
import typing as t
import logging
import importlib.metadata  # python >= 3.8

from .version import CheckVersion
from .utils import version_to_tuple
from ..config import Parameter

__all__ = ("CheckPythonPackage",)

logger = logging.getLogger(__name__)


class CheckPythonPackage(CheckVersion):
    """Check the availability and version of a python package"""

    # The message for checking python packages
    msg = Parameter(
        "CHECKPYTHONPACKAGE.MSG",
        default="Check python package '{check.raw_value}'",
    )

    aliases = ("checkPythonPackage", "checkPythonPkg", "CheckPythonPkg")

    def get_current_version(self) -> t.Union[None, t.Tuple[int]]:
        # Get the package name, operator and version to check against (the last
        # 2 aren't used here)
        pkg_name, op, version = self.value

        # Nothing to do if the package name was not found
        if pkg_name is None:
            return None

        # Method 1 -- importlib.metadata
        try:
            # Returns a version string--e.g. '0.9.3'
            version_string = importlib.metadata.version(pkg_name)

            # Return the version tuple--e.g. (0, 9, 3)
            return version_to_tuple(version_string) if version_string else None
        except importlib.metadata.PackageNotFoundError:
            return None
